         
syntax = "proto2";
package NV.Profiler;
enum ProfilerCategory
{
    // Every feature counts Category 0 as invalid. Then the array of categories.size() is always >  1.
    CategoryInvalid = 0;
    CategoryConfiguration = 1;
    CategoryControl = 2;
    CategoryResults = 3;
    CategoryStatus = 4;
}
